module abstract_simulator
  
  use abstract_vector
  use abstract_gradient
  use abstract_inner_product
  implicit none

  private
  public :: simulator_class

  ! A generic simulator for use with the generic conjugate gradient solver.

  ! SIMULATOR_CLASS defines the following method:
  !   eval : (deferred) Takes a vector x and an inner product ip, and
  !          returns a gradient, with respect to ip, of some function
  !          evaluated at x.

  type, abstract :: simulator_class
    private
  contains
    private
    procedure(s4d), public, deferred :: eval
  end type simulator_class

  abstract interface
    subroutine s4d (this,x,g,ip)
      import simulator_class
      import vector_class
      import gradient_class
      import inner_product_class
      class(simulator_class),     intent(in) :: this
      class(vector_class),        intent(in) :: x
      class(gradient_class),      intent(inout) :: g
      class(inner_product_class), intent(inout) :: ip
    end subroutine
  end interface
end module abstract_simulator
