module concrete_gradient
  use abstract_vector
  use abstract_gradient
  use concrete_vector, only : trivial_vector_type
!!  use concrete_inner_product, only : trivial_inner_product_type
  implicit none

  ! TRIVIAL_GRADIENT_TYPE provides a simple concrete example of the abstract
  ! type defined by GRADIENT_CLASS. Other definitions are possible. The
  ! abstract conjugate gradients routine will work with all of them!

  type, extends(gradient_class) :: trivial_gradient_type
    real :: elements(100)
  contains
    procedure :: assign      => my_assign
    procedure :: negate      => my_negate
    procedure :: subtract    => my_subtract
    procedure :: add         => my_add
    procedure :: multiply_by => my_multiply_by
    procedure :: saxpy       => my_saxpy
    procedure :: to_vector   => my_to_vector
  end type trivial_gradient_type

contains

  subroutine my_assign (this,v)
    class(trivial_gradient_type), intent(inout) :: this
    class(vector_class),          intent(in)    :: v

    select type (v)
    type is (trivial_gradient_type)
      this%elements(:) = v%elements(:)
    end select
!    call this%set_inner_product(v%inquire_inner_product())
  end subroutine my_assign

  subroutine my_negate (this)
    class(trivial_gradient_type), intent(inout) :: this

    this%elements(:) = -this%elements(:)
  end subroutine my_negate

  subroutine my_subtract (this,v)
    class(trivial_gradient_type), intent(inout) :: this
    class(vector_class), intent(in)    :: v
!!    type(trivial_inner_product_type), pointer  :: p1, p2

!!    p1 => this%inquire_inner_product()
!!    p2 =>    v%inquire_inner_product()
!!    if (.not.associated(p1,target=p2)) then
!!      call abor1 ('subtract gradients with incompatible inner products')
!!    endif
    select type (v)
    type is (trivial_gradient_type)
      this%elements(:)= this%elements(:) - v%elements(:)
    end select
  end subroutine my_subtract

  subroutine my_add (this,v)
    class(trivial_gradient_type), intent(inout) :: this
    class(vector_class), intent(in)    :: v
!!    type(trivial_inner_product_type), pointer  :: p1, p2

!!    p1 => this%inquire_inner_product()
!!    p2 =>    v%inquire_inner_product()
!!    if (.not.associated(p1,target=p2)) then
!!      call abor1 ('add gradients with incompatible inner products')
!!    endif
    select type (v)
    type is (trivial_gradient_type)
      this%elements = this%elements(:) + v%elements(:)
    end select
  end subroutine my_add

  subroutine my_multiply_by (this,s)
    class(trivial_gradient_type), intent(inout) :: this
    real, intent(in)                            :: s

    this%elements = s*this%elements(:)
  end subroutine my_multiply_by

  subroutine my_saxpy (this,s,v)
    class(trivial_gradient_type), intent(inout) :: this
    real, intent(in)                            :: s
    class(vector_class), intent(in)    :: v
!!    type(trivial_inner_product_type), pointer  :: p1, p2

!!    p1 => this%inquire_inner_product()
!!    p2 =>    v%inquire_inner_product()
!!    if (.not.associated(p1,target=p2)) then
!!      call abor1 ('saxpy gradients with incompatible inner products')
!!    endif
    select type (v)
    type is (trivial_gradient_type)
      this%elements = this%elements(:) + s*v%elements(:)
    end select
  end subroutine my_saxpy

  subroutine my_to_vector (this,v)
    class(trivial_gradient_type), intent(in)    :: this
    class(vector_class),          intent(inout) :: v

    select type (v)
    type is (trivial_vector_type)
      v%elements = this%elements
    end select
  end subroutine my_to_vector

end module concrete_gradient
