MODULE mo_linked_list

  ! finalzer instead of delete_list, solves scoping
  ! same_type_as() for checking type coherence

  TYPE, PRIVATE :: list_item
    CLASS(*), POINTER :: item_of_list
    TYPE(list_item), POINTER :: next => NULL()
  END TYPE list_item

  TYPE, PUBLIC :: list
    PRIVATE
    TYPE(list_item), POINTER :: start => NULL()
    TYPE(list_item), POINTER :: end   => NULL()
  CONTAINS
    PROCEDURE, NON_OVERRIDABLE :: add_list_item
    PROCEDURE, NON_OVERRIDABLE :: delete_list
  END TYPE list

  TYPE, EXTENDS(list), PUBLIC :: list_iterator
    TYPE(list_item), POINTER :: list_start => NULL()
    TYPE(list_item), POINTER :: current => NULL()
  CONTAINS
    PROCEDURE, NON_OVERRIDABLE :: set => set_list
    PROCEDURE, NON_OVERRIDABLE :: iterate
  END TYPE list_iterator

CONTAINS

  SUBROUTINE set_list(this, root)
    CLASS(list_iterator), INTENT(inout) :: this
    CLASS(list), INTENT(in) :: root
    IF (ASSOCIATED(root%start)) THEN
      this%list_start => root%start
      this%current    => root%start
    ENDIF
  END SUBROUTINE set_list

  FUNCTION iterate(this) RESULT(next)
    CLASS(list_iterator), INTENT(inout) :: this
    CLASS(*), POINTER :: next 
    IF (.NOT. ASSOCIATED(this%list_start)) THEN
      WRITE (0,*) 'Iterator list not initialized!'
      next => NULL()
    ELSE
      IF (ASSOCIATED(this%current)) THEN
        next => this%current%item_of_list
        this%current => this%current%next
      ELSE
        next => NULL()
      ENDIF
    ENDIF
  END FUNCTION iterate

  SUBROUTINE add_list_item(this, item)
    CLASS(list), INTENT(inout) :: this
    CLASS(*), INTENT(in) :: item
    IF (ASSOCIATED(this%start)) THEN
      ALLOCATE(this%end%next)
      ALLOCATE(this%end%next%item_of_list, source = item)
      this%end => this%end%next
    ELSE
      ALLOCATE(this%start)
      ALLOCATE(this%start%item_of_list, source = item)
      this%end => this%start
    ENDIF
  END SUBROUTINE add_list_item

  RECURSIVE SUBROUTINE delete_list(this)
    CLASS(list), INTENT(inout) :: this
    TYPE(list_item), POINTER :: tmp
    IF (ASSOCIATED(this%start)) THEN
      tmp => this%start%next
      DEALLOCATE(this%start)
      this%start => tmp
    ENDIF
    IF (ASSOCIATED(this%start)) &
         CALL delete_list(this)
  END SUBROUTINE delete_list

END MODULE mo_linked_list

PROGRAM linked_list_test

  USE mo_linked_list

  TYPE :: metadata
    INTEGER :: variable = 31
  END TYPE metadata

  TYPE :: stream_element
    TYPE(metadata) :: info
    REAL, ALLOCATABLE :: data(:)
  END TYPE stream_element

  TYPE(list) :: io_list
  TYPE(list_iterator) :: io

  TYPE(stream_element), TARGET :: item1(12)

  CLASS(*), POINTER :: io1
  INTEGER :: i

  DO i = 1, 12
    item1(i)%info%variable = i
    CALL add_list_item(io_list, item1(i))
  ENDDO
  
  CALL io%set(io_list)
  DO
    io1 => io%iterate()
    IF (.NOT. ASSOCIATED(io1)) EXIT
    SELECT TYPE(io1)
    TYPE IS (stream_element)
      WRITE (0,'(a,i3)') 'Variable: ', io1%info%variable    
    END SELECT
  ENDDO

  CALL io_list%delete_list

END PROGRAM
